/*
 * Hash tables
 *
 *
 *
 * We have 3 hash types
 *
 * SCM_HASH_T_GEN, the general one, SCM_HASH_T_SYMBOL and SCM_HASH_T_ATOM.
 *
 *
 * thread of SCM_HASH_T_GEN is: ( (key . value) ... )
 *
 * thread -><pair>[ . | *-]--->...
 *                  |
 *                  v
 *			<pair>[ <any>key | <any>value ]
 *
 *
 * Thread of SCM_HASH_T_SYMBOL is: ( sym ... )
 *
 * thread ->  <pair> [ . | *-]-->
 *                     |
 *                     v
 *            <sym>  [ <atom>name | <any>value ]
 *
 *
 * Thread of SCM_HASH_T_ATOM is: 
 *
 * thread -> <atom> [ <cstr>name | *-]---> ...
 *
 */

#include "s.h"

/****************************************************************
 * Type functions
 ****************************************************************/

void scm_hash_mark(SOBJ obj)
{
  int i;
  SOBJ l;
  SCM_Hash *h = SCM_HASH(obj);

  for (i = 0; i < h->hsize; i++) {
	l = h->hash[i];
	if (l != NULL) scm_gc_mark(l);
  }
}

void scm_hash_sweep(SOBJ obj)
{
  if (SCM_HASH(obj)) scm_free(SCM_HASH(obj));
  SCM_HASH(obj) = NULL;
}

static void scm_hash_dump(SOBJ obj, PORT *p, int flag)
{
  int i;
  SCM_Hash *h = SCM_HASH(obj);
  SOBJ l;

  switch(h->type) {
  case SCM_HASH_T_GEN:		port_puts(p, "#{");   				break;
  case SCM_HASH_T_SYMBOL:	port_puts(p, "#{<symbol-hash> ");  	break;
  case SCM_HASH_T_ATOM:		port_puts(p, "#{<atom-hash> "); 	break;
  default:					port_puts(p, "#{<???> ");
  }
  
  for (i = 0; i < h->hsize; i++) {
	l = h->hash[i];
	while(l) {
	  switch(h->type) {
	  case SCM_HASH_T_ATOM:
		scm_write_obj(l,p,flag);
		l = SCM_ATOM_NEXT(l);
		break;
	  case SCM_HASH_T_SYMBOL:
	  case SCM_HASH_T_GEN:
		scm_write_obj(SCM_CAR(l), p, flag);
		l = SCM_CDR(l);
		break;
	  }
	  port_putc(p, ' ');
	}
	/* This works for debugging hash threads...
	  if (h->hash[i]) scm_write_obj(h->hash[i], p, flag);
	*/
  }
  port_putc(p, '}');
}

/****************************************************************
 * Helper functions
 ****************************************************************/
Uint scm_hash_string(char *str)
{
  int c;
  unsigned int r = 0;
  while((c = *str++) != 0) {
	r = (r*7) ^ c;
	/*	r += (r << 3) + c; */
  }
  return(r);
}

Uint scm_hash_string1(char *str)
{
  unsigned int r = 0, c, x;

  while( (c = *str++) != 0) {
	x = (r << 4) + c;
	if (x) 	r = (r ^ (x % ((~0U >> 4U) + 1))) & (~0U >> 4U);
  }
  return(r);
}

Uint scm_hash_pointer(void *ptr)
{
  long n = (long) ptr;
  return( (n>>1) * 1103515245 );
}

Uint scm_hash_code(SCM_Hash *h, SOBJ key)
{
  int hc;
  
  switch(SCM_OBJTYPE(key)) {
  case SOBJ_T_ATOM:
	hc = scm_hash_string(SCM_ATOM_NAME(key));
	break;
	
  case SOBJ_T_STRING:
	hc = scm_hash_string(SCM_STR_VALUE(key));
 	break;
	
  case SOBJ_T_SYMBOL:
	hc = scm_hash_string(SCM_ATOM_NAME(SCM_SYM_NAME(key)));
	break;

  default:
	hc = scm_hash_pointer(key);
  }
#ifdef DEBUG
  scm_puts("; hash code for "); scm_cdisplay(key);
  scm_puts("="); scm_putn(hc);  scm_puts("\n");
#endif
  return(hc % h->hsize);
}

void scm_hash_print(SOBJ obj, PORT *p)
{
  scm_hash_dump(obj, p, TRUE);
}

void scm_hash_write(SOBJ obj, PORT *p)
{
  scm_hash_dump(obj, p, FALSE);
}

/****************************************************************
 * Scheme functions
 ****************************************************************/
SOBJ scm_mkhash(int type) 
{
  SOBJ new;
  SCM_Hash *h;
  int i;
  
  new = scm_newcell(SOBJ_T_HASH);
  h   = scm_must_alloc( sizeof(SCM_Hash) );

  h->hsize = 4;
  h->type  = type;
  h->hash  = scm_must_alloc(sizeof(SOBJ) * h->hsize);
  h->nkeys = 0;
  h->maxkeys = h->hsize * SCM_MAX_HASH_DEPTH;
  for (i = 0; i < h->hsize; i++) {
	h->hash[i] = NULL;
  }
  SCM_HASH(new) = h;
  return(new);
}

SOBJ scm_make_hash(SOBJ type)
{
  int n;

  if (!SCM_INUMP(type)) 	err("make-hash: bad hash type", type);
  n = SCM_INUM(type);
  if (n != SCM_HASH_T_GEN && n != SCM_HASH_T_SYMBOL && n != SCM_HASH_T_ATOM) {
	err("make-hash: bad hash-type", type);
  }
  return(scm_mkhash(n));
}

/*-- principle is:
 * - build an array containing ref to all nodes of hash
 * - alloc a new hash[]
 * - fill the new hash[]
 * - free the ref array
 */
void scm_rebuild_hash(SOBJ hash)
{
  SOBJ *ref, l;
  int   nr, i, hc;
  SCM_Hash *h;

  h = SCM_HASH(hash);

  /* collect all atoms to ref[] */
  ref = scm_must_alloc(sizeof(SOBJ) * h->nkeys);
  nr = 0;
  for (i = 0; i < h->hsize; i++) {
	for (l = h->hash[i]; l; l = SCM_ATOM_NEXT(l)) {
	  ref[nr++] = l;
	}
  }

  /* resize the hash */
  scm_free(h->hash);
  h->hsize = h->hsize*4;
  h->maxkeys = h->hsize * SCM_MAX_HASH_DEPTH;
  h->hash = scm_must_alloc(sizeof(SOBJ) * h->hsize);
  
  /* initialize to nil */
  for (i = 0; i < h->hsize; i++) 	h->hash[i] = NULL;

  switch(h->type) {
  case SCM_HASH_T_GEN:
	for (i = 0; i < nr; i++) {
	  hc = scm_hash_code(h, SCM_CAAR(ref[i]));
	  SCM_CDR(ref[i]) = h->hash[hc];
	  h->hash[hc] = ref[i];
	}
	break;

  case SCM_HASH_T_SYMBOL:
	for (i = 0; i < nr; i++) {
	  hc = scm_hash_code(h, SCM_CAAR(ref[i]));
	  SCM_CDR(ref[i]) = h->hash[hc];
	  h->hash[hc] = ref[i];
	}
	break;

  case SCM_HASH_T_ATOM:
	for (i = 0; i < nr; i++) {
	  hc = scm_hash_string(SCM_ATOM_NAME(ref[i])) % h->hsize;
	  SCM_ATOM_NEXT(ref[i]) = h->hash[hc];
	  h->hash[hc] = ref[i];
	}
  }
  scm_free(ref);
}

/* KKK implement type dependant hash */

/* search a hash for the key.
 * returns a pointer to the toplevel node or NULL if not found. Toplevel
 * node is:
 * - the atom itself in case of atom hash
 * - the pointer to a pair which car is a symbol in case of symbol hash
 * - the pointer to a pair pointing to a key/value pair otherwise
 */
SOBJ scm_hash_search_key(SOBJ hash, SOBJ key, int *hcode, char **key_str)
{
  SCM_Hash *h;
  int i;
  SOBJ l;
  char *p;
  
  h = SCM_HASH(hash);
  p = NULL;
  
  switch(h->type) {

  case SCM_HASH_T_ATOM:
	
	switch(SCM_OBJTYPE(key)) {
	case SOBJ_T_ATOM:	p = SCM_ATOM_NAME(key); 				break;
	case SOBJ_T_STRING:	p = SCM_STR_VALUE(key);					break;
	case SOBJ_T_SYMBOL: p = SCM_ATOM_NAME(SCM_SYM_NAME(key));	break;
	default:
	  err("hash-search: atom-hash key must be atom|string|symbol",key);
	}
	i = scm_hash_string(p) % h->hsize;
	*hcode = i;  *key_str = p;
	for (l = h->hash[i]; l; l = SCM_ATOM_NEXT(l)) {
	  if (strcmp(p, SCM_ATOM_NAME(l)) == 0)	return(l);
	}
	break;
	
  case SCM_HASH_T_SYMBOL:
	
	switch(SCM_OBJTYPE(key)) {
	case SOBJ_T_ATOM:	p = SCM_ATOM_NAME(key);					break;
	case SOBJ_T_STRING:	p = SCM_STR_VALUE(key);					break;
	case SOBJ_T_SYMBOL:	p = SCM_ATOM_NAME(SCM_SYM_NAME(key));	break;
	default:
	  err("hash-search: symbol-hash key must be atom|string|symbol",key);
	}
	i = scm_hash_string(p) % h->hsize;
	*hcode = i;  *key_str = p;
	for (l = h->hash[i]; l; l = SCM_CDR(l)) {
	  if (streq(p, SCM_ATOM_NAME(SCM_SYM_NAME(SCM_CAR(l)))))	return(l);
	}
	break;
	
  case SCM_HASH_T_GEN:
	
	i = scm_hash_code(h, key);
	*hcode = i;
	*key_str = NULL;
	for (l = h->hash[i]; l; l = SCM_CDR(l)) {
	  if (scm_equal(key, SCM_CAAR(l)) != scm_false) return(l);
	}
	break;

  default:
	err("hash-search: illegal hash-type", hash);
  }
  return(NULL);
}


SOBJ scm_hash_set(SOBJ hash, SOBJ key, SOBJ value)
{
  SOBJ node;
  SCM_Hash *h;
  int hc;						/* the hash code */
  char *keystr;					/* the key string used to search */

  if (SCM_OBJTYPE(hash) != SOBJ_T_HASH) err("hash-set!: bad hash", hash);

  h = SCM_HASH(hash);
  node = scm_hash_search_key(hash, key, &hc, &keystr);

  if (node) {					/* found */
	switch(h->type) {
	case SCM_HASH_T_ATOM:	/* atom hash have no value */			break;
	case SCM_HASH_T_SYMBOL:	SCM_SYM_VALUE(SCM_CAR(node)) = value; 	break;
	case SCM_HASH_T_GEN:	SCM_CDR(SCM_CAR(node)) = value;			break;
	}
	return(node);
  }

  /* not found, create the node */
  
  switch(h->type) {
  case SCM_HASH_T_ATOM:
	{
	  SOBJ new = scm_newcell(SOBJ_T_ATOM);
	  SCM_ATOM_NAME(new) = scm_must_strdup(keystr);
	  SCM_ATOM_NEXT(new) = h->hash[hc];
	  h->hash[hc] = new;
	  node = new;
	  break;
	}
  case SCM_HASH_T_SYMBOL:
	{
	  SOBJ sym = scm_newcell(SOBJ_T_SYMBOL);
	  SCM_SYM_NAME(sym) = scm_mkatom(keystr);
	  SCM_SYM_VALUE(sym) = value;
	  h->hash[hc] = scm_cons(sym, h->hash[hc]);
	  node = h->hash[hc];
	  break;
	}
  case SCM_HASH_T_GEN:
	{
	  SOBJ assoc = scm_cons(key, value);
	  h->hash[hc] = scm_cons(assoc, h->hash[hc]);
	  node = h->hash[hc];
	  break;
	}
  default:
	err("hash-set!: illegal hash type", hash);
  }
  
  h->nkeys++;
  if (h->nkeys >= h->maxkeys) { scm_rebuild_hash(hash); }
  return(node);
}

SOBJ scm_hash_ref(SOBJ hash, SOBJ key)
{
  SOBJ node;
  SCM_Hash *h;
  int hc;						/* the hash code */
  char *keystr;					/* the key string used to search */

  if (SCM_OBJTYPE(hash) != SOBJ_T_HASH) err("hash-ref: bad hash", hash);

  h = SCM_HASH(hash);
  node = scm_hash_search_key(hash, key, &hc, &keystr);
  if (node) {
	switch(h->type) {
	case SCM_HASH_T_SYMBOL:	return(SCM_SYM_VALUE(SCM_CAR(node)));
	case SCM_HASH_T_GEN:	return(SCM_CDR(SCM_CAR(node)));
	}
	return(node);
  }
  return(scm_undefined);
}

/* just like hash-ref, but returns the assoc node, not the value */

SOBJ scm_hash_search(SOBJ hash, SOBJ key)
{
  SOBJ node;
  SCM_Hash *h;
  int hc;						/* the hash code */
  char *keystr;					/* the key string used to search */

  if (SCM_OBJTYPE(hash) != SOBJ_T_HASH) err("hash-set!: bad hash", hash);

  h = SCM_HASH(hash);
  node = scm_hash_search_key(hash, key, &hc, &keystr);
  if (node) {
	switch(h->type) {
	case SCM_HASH_T_SYMBOL:
	case SCM_HASH_T_GEN:	return(SCM_CAR(node));
	}
	return(node);
  }
  return(scm_undefined);
}


SOBJ scm_hash_to_list(SOBJ hash)
{
  SOBJ l, p;
  int i;
  SCM_Hash *h;
  
  if (!SCM_HASHP(hash)) err("hash->list: bad hash", hash);

  l = NULL;
  h = SCM_HASH(hash);
  for (i = 0; i < h->hsize; i++) {
	for (p = h->hash[i]; p; p = SCM_CDR(p)) {
	  l = scm_cons(SCM_CAR(p), l);
	}
  }
  return(l);
}

SOBJ scm_list_to_hash(SOBJ l)
{
  SOBJ hash;
  SOBJ e;
  
  if (!SCM_PAIRP(l))	err("list->hash: bad list", l);
  
  hash =  scm_make_hash(SCM_MKINUM(SCM_HASH_T_GEN));
  while(l && SCM_PAIRP(l)) {
	e = SCM_CAR(l);
	if (!SCM_PAIRP(e)) 	err("list->hash: bad entry", e);
	scm_hash_set(hash, SCM_CAR(e), SCM_CDR(e));
	l = SCM_CDR(l);
  }
  return(hash);
}

SOBJ scm_hash_stat(SOBJ hash)
{
  SCM_Hash *h;
  SOBJ l;
  int i, n, max;
  int depth[10];

  if (!SCM_HASHP(hash)) err("hash-stat: bad hash", hash);
	
  h = SCM_HASH(hash);
  scm_puts("hash statistics:\n");
  scm_puts("hsize: ");  	scm_putn(h->hsize);  	scm_putc('\n');
  scm_puts("nkeys: ");  	scm_putn(h->nkeys);  	scm_putc('\n');
  scm_puts("maxkeys: "); 	scm_putn(h->maxkeys);	scm_putc('\n');
  
  scm_puts("distribution: ");
  max = 0;
  for (i = 0; i < 10; i++) { depth[i] = 0; }
  
  for (i = 0; i < h->hsize; i++) {
	n = 0;
	for (l = h->hash[i]; l; l = SCM_CDR(l)) n++;
	if (n > max) max = n;
	depth[ (n < 10) ? n : 9 ]++;
	scm_putn(n); scm_putc(' ');
  }
  scm_putc('\n');
  scm_puts("# keys in # thread\n");
  for (i = 0; i < 10; i++) {
	scm_putn(i); scm_puts(" "); scm_putn(depth[i]); scm_putc('\n');
  }
  scm_puts("worst case: "); scm_putn(max);	scm_putc('\n');

  return(scm_undefined);
}

SOBJ scm_hashp(SOBJ x)
{
  return(SCM_MKBOOL(SCM_OBJTYPE(x) == SOBJ_T_HASH));
}

SOBJ scm_atom_hashp(SOBJ x)
{
  return(SCM_MKBOOL( (SCM_OBJTYPE(x) == SOBJ_T_HASH) && (SCM_HASH(x)->type == SCM_HASH_T_ATOM)));
}

SOBJ scm_symbol_hashp(SOBJ x)
{
  return(SCM_MKBOOL( (SCM_OBJTYPE(x) == SOBJ_T_HASH) && (SCM_HASH(x)->type == SCM_HASH_T_SYMBOL)));
}

SOBJ scm_gen_hashp(SOBJ x)
{
  return(SCM_MKBOOL( (SCM_OBJTYPE(x) == SOBJ_T_HASH) && (SCM_HASH(x)->type == SCM_HASH_T_GEN)));
}

void scm_init_hash()
{
  scm_add_cprim("make-hash", 	scm_make_hash,		1);
  scm_add_cprim("hash-set!",	scm_hash_set,		3);
  scm_add_cprim("hash-ref",		scm_hash_ref,		2);
  scm_add_cprim("hash->list",	scm_hash_to_list,	1);
  scm_add_cprim("list->hash",	scm_list_to_hash,	1);
  scm_add_cprim("hash-stat",	scm_hash_stat,		1);
  scm_add_cprim("hash?",		scm_hashp,			1);
  scm_add_cprim("atom-hash?",		scm_atom_hashp,		1);
  scm_add_cprim("symbol-hash?",		scm_symbol_hashp,	1);
  scm_add_cprim("generic-hash?",	scm_gen_hashp,		1);
}
